;;                       PERSONAJES TOLKIENIANOS
;;                        Marc Meléndez "Bruka"
;;
;;    Este programa intenta adivinar un personaje tolkieniano
;;    que piensa el usuario. Si no lo conoce, aprende el nuevo
;;    personaje. El interés de este programa es meramente
;;    didáctico.
;;
;;    (c) Marc Meléndez Schofield, 2006. Se permite la copia,
;;    modificación y redistribución de este programa para fines
;;    no lucrativos.
;;
;;    A los que lean el programa: perdón por mi acento en Scheme,
;;    tiendo a escribir los paréntesis como bloques al estilo de C,
;;    porque me resulta más claro. Perdón también por mi tendencia
;;    a nombrar las cosas en inglés (como las instrucciones son
;;    palabras inglesas, tiendo a pensar el programa en inglés).
;;    En este caso, intenté nombrar las rutinas y variables en
;;    español (de ahí que haya una variable 'arbol', una rutina
;;    'quien_es', etc), pero la costumbre me llevó a nombrar muchas
;;    otras cosas en inglés.
;;    En la versión del programa para DrScheme en windows, me he visto
;;    obligado a eliminar las tildes y las "ñ", porque este Scheme
;;    parece no reconocerlas.

;;  ESTRUCTURA DE DATOS EN ÁRBOL
;;
;;  Los datos de los personajes están recursivamente organizados en
;;  árbol de esta manera:
;;                        (diferencia rama rama)
;;  donde cada rama puede ser o bien de la forma (diferencia rama rama)
;;  o bien una hoja:
;;                        (nombre_del_personaje)
;;  El programa simplemente recorre el árbol preguntando la diferencia
;;  y escogiendo la primera rama en caso de que la respuesta del
;;  usuario sea afirmativa y la segunda en caso contrario. Cuando llega
;;  a una hoja, pregunta si el personaje que eligió el usuario es el
;;  nombrado en la hoja. En caso afirmativo, el programa ha acertado y
;;  termina. Si no ha acertado, pregunta por el nuevo personaje y por
;;  lo que lo diferencia del anterior, y sustituye la hoja por
;;  (nueva_diferencia (nombre_del_nuevo_personaje) (nombre_del_personaje))
;;  Luego guarda los datos en el fichero.
;;  La idea general quedará más clara con un ejemplo. Utilizaremos un
;;  árbol sencillo:
;;                    Pertenece a los pueblos libres
;;                      Sí _________|__________ No
;;                         |                  |
;;                    Es un enano       Boca de Sauron
;;                Sí ______|______ No
;;                   |           |
;;                 Gimli      Trancos
;;  El árbol quedaría así:
;;  ("Pertenece a los pueblos libres" ("Es un Enano" ("Gimli") ("Trancos))
;;     ("Boca de Sauron"))
;;  El programa pregunta "¿Pertenece a los pueblos libres?". Si el usuario
;;  responde que sí, pregunta "¿Es un enano?". Si responden que no, pregunta
;;  "¿Es Trancos?". Si responden "sí", el programa escribe "¡Acerté!" y 
;;  termina. Pero supongamos que el usuario ha pensado en Légolas. En este
;;  caso no ha acertado, así que pregunta al usuario su personaje y qué
;;  diferencia a Légolas de Trancos. Imaginemos que el usuario responde
;;  "Es un elfo". El programa cambiará el árbol de modo que quede así:
;;                    Pertenece a los pueblos libres
;;                      Sí _________|__________ No
;;                         |                  |
;;                    Es un enano       Boca de Sauron
;;                Sí ______|______ No
;;                   |           |
;;                 Gimli     Es un elfo
;;                      Sí ______|______ No
;;                         |           |
;;                      Légolas      Trancos
;;
;;  CÓDIGO EN SCHEME
;;
;;  Lo primero es abrir el fichero donde se guardan los personajes y
;;  sus características. Estos datos están en el fichero "personajes.dat".
;;  Los datos se guardan en la variable 'arbol' y se cierra el fichero.

(define port (open-input-file "personajes.dat"))
(define arbol (read port))
(close-input-port port)

;;  La rutina 'get-string' toma las palabras que introduce el usuario con el
;;  teclado y las convierte en una hilera de caracteres de texto (string).

(define get-string
  (lambda ()
    (let ((input-symbol (read)))
      (if (eof-object? input-symbol)
          ""
          (if (eq? input-symbol 'fin)
	    ""
	    (string-append (symbol->string input-symbol) " " (get-string))
          )
      )
    )
  )
)

;;  'string-cap' sirve sólamente para eliminar el espacio extra que
;;  aparece al final del texto que devuelve 'get-string'.

(define string-cap
  (lambda (str)
    (substring str 0 (- (string-length str) 1))
  )
)

;;  La rutina 'si?' se utiliza para identificar las respuestas
;;  afirmativas (ya que algunas versiones de Scheme considerarán que
;;  'Si' y 'si' son signos distintos). También se puede responder 's'.
;;  Cualquier otra respuesta se considera negativa.

(define si?
  (lambda ()
    (let ((answer (read)))
      (if (or (eq? answer 'sí) (eq? answer 's) (eq? answer 'si)
	      (eq? answer 'Sí) (eq? answer 'S) (eq? answer 'SI))
	  #t
	  #f
      )
    )
  )
)

;;  La próxima función se ejecuta sólo cuando el programa no consigue
;;  adivinar el personaje del usuario. Utiliza 'get-string' y 
;;  'string-cap' para conseguir los datos introducidos por el usuario.
;;  Después sustituye la hoja con el antiguo personaje por la
;;  estructura (diferencia (nuevo_personaje) (antiguo_personaje))
;;  tal como se explica más arriba en la sección "estructura en
;;  árbol". 'quien_es' toma como argumento una hoja del arbol de datos.

(define quien_es
  (lambda (arbol)
    (let
      (
       (nom "")
       (dif "")
      )
      (display "Entonces... ¿Quién es? ")
;      (display " (Cuando termines, pulsa [Enter] y después escribe fin y pulsa [Enter])")
      (newline)
      (set! nom (string-cap (get-string)))
      (display "Ahh...Por favor, escribe algo que sea falso para ")
      (display (car arbol))
      (display " pero verdadero para ")
      (display nom)
      (display ". (Cuando termines, pulsa [Enter] y después escribe fin y pulsa [Enter])")
      (newline)
      (set! dif (string-cap (get-string)))
      (cons dif (cons (cons nom '()) (cons arbol '())))
    )
  )
)

;;  'juego' es la rutina que recorre el árbol de los personajes. Es
;;  una rutina recursiva. Toma como argumento un arbol.
;;  Comienza preguntando la diferencia. Si el usuario responde "Sí",
;;  se ejecuta 'juego' con la primera rama como argumento, si la 
;;  respuesta es "no" se ejecuta 'juego' con la segunda rama como
;;  argumento. Si el argumento de 'juego' resulta ser una hoja, el
;;  programa simplemente pregunta si el personaje es el que hay
;;  almacenado en la hoja. Si acierta, el programa ha terminado. Si
;;  falla, utiliza la rutina 'quien_es' para actualizar el árbol.
;;  La rutina tiene la típica estructura recursiva de Scheme, por lo que
;;  la última parte de la rutina se utiliza para reconstruir
;;  recursivamente el árbol de datos.

(define juego
  (lambda (arbol)
    (if (equal? (length arbol) 1)
      (begin
        (display "¿Es ")
        (display (car arbol))
        (display "? ")
        (if (si?)
          (begin
            (display "¡Acerté!")
            arbol
          )
          (quien_es arbol)
        )
      )
      (begin
        (display "¿")
        (display (car arbol))
        (display "? ")
        (if (si?)
          (cons (car arbol) 
                (cons (juego (cadr arbol))
                      (cons (caddr arbol) '())
                )
          )
          (cons (car arbol)
                (cons (cadr arbol)
                      (cons (juego (caddr arbol)) '())
                )
          )
        )
      )
    )
  )
)

;;  La rutina 'init' pone en marcha el programa y guarda el árbol
;;  actualizado (en caso de que haya cambiado) en el fichero
;;  "personajes.dat". Para la versión de DrScheme, he tenido que
;;  incluir una instrucción que borre el fichero de datos antes
;;  de intentar grabar en él, ya que la instrucción 'open-output-
;;  file' devuelve un error si el archivo ya existe.

(define init
  (lambda ()
    (begin
      (delete-file "personajes.dat")
      (set! port (open-output-file "personajes.dat"))
      (newline) (newline)
      (display "     ***************************************") (newline)
      (display "     *                                     *") (newline)
      (display "     *    PERSONAJES DE LA TIERRA MEDIA    *") (newline)
      (display "     *                                     *") (newline)
      (display "     ***************************************") (newline)
      (newline)
      (display "Piensa en un personaje de la Tierra Media.") (newline)
      (set! arbol (juego arbol))
      (write arbol port)
      (close-output-port port)
      (newline)
      (display "Gracias por jugar. Para volver a jugar, ejecuta el comando (init)") (newline)
    )
  )
)

; Finalmente ejecutamos 'init'.

(init)
